/* Non-ASCII keycodes */

#ifndef _KEYCODES_H_
#define _KEYCODES_H_

#define K_ESC     0x1B
#define K_CAPSLK  0x80
#define K_F1      0x81
#define K_F2      0x82
#define K_F3      0x83
#define K_F4      0x84
#define K_F5      0x85
#define K_F6      0x86
#define K_F7      0x87
#define K_F8      0x88
#define K_F9      0x89
#define K_F10     0x8A
#define K_F11     0x8B
#define K_F12     0x8C
#define K_UP      0x8D
#define K_LEFT    0x8E
#define K_DOWN    0x8F
#define K_RIGHT   0x90
#define K_INS     0x91
#define K_DEL     0x92
#define K_HOME    0x93
#define K_END     0x94
#define K_PGUP    0x95
#define K_PGDN    0x96
#define K_NUMLK   0x97
#define K_SCRLK   0x98
#define K_PRTSC   0x99
#define K_BREAK   0x9A

#endif
